"""
The simple_travel_example.py file from the Pyhop distribution, with some
minor changes to make it compatible with GTPyhop:
  - declare a domain to hold the action and method definitions
  - replace all references to pyhop with gtpyhop
  - change 'verbose' to a global variable instead of a keyword argument

To keep this file as close as possible to the Pyhop version, it doesn't use
the test harness that's used with the other example files.
-- Dana Nau <nau@umd.edu>, July 20, 2021
"""

import gtpyhop


# declare a domain to hold the action and method definitions
#
gtpyhop.Domain('pyhop_simple_travel_example')

def taxi_rate(dist):
    return (1.5 + 0.5 * dist)

def walk(state,a,x,y):
    if state.loc[a] == x:
        state.loc[a] = y
        return state
    else: return False

def call_taxi(state,a,x):
    state.loc['taxi'] = x
    return state
    
def ride_taxi(state,a,x,y):
    if state.loc['taxi']==x and state.loc[a]==x:
        state.loc['taxi'] = y
        state.loc[a] = y
        state.owe[a] = taxi_rate(state.dist[x][y])
        return state
    else: return False

def pay_driver(state,a):
    if state.cash[a] >= state.owe[a]:
        state.cash[a] = state.cash[a] - state.owe[a]
        state.owe[a] = 0
        return state
    else: return False

gtpyhop.declare_operators(walk, call_taxi, ride_taxi, pay_driver)
print('')
gtpyhop.print_operators()



def travel_by_foot(state,a,x,y):
    if state.dist[x][y] <= 2:
        return [('walk',a,x,y)]
    return False

def travel_by_taxi(state,a,x,y):
    if state.cash[a] >= taxi_rate(state.dist[x][y]):
        return [('call_taxi',a,x), ('ride_taxi',a,x,y), ('pay_driver',a)]
    return False

gtpyhop.declare_methods('travel',travel_by_foot,travel_by_taxi)
print('')
gtpyhop.print_methods()

state1 = gtpyhop.State('state1')
state1.loc = {'me':'home'}
state1.cash = {'me':20}
state1.owe = {'me':0}
state1.dist = {'home':{'park':8}, 'park':{'home':8}}

print("""
********************************************************************************
Call gtpyhop.pyhop(state1,[('travel','me','home','park')]) with different verbosity levels
********************************************************************************
""")

print("- If verbose=0, GTPyhop returns the solution but prints nothing.\n")
gtpyhop.verbose = 0
gtpyhop.pyhop(state1,[('travel','me','home','park')])

print('- If verbose=1, GTPyhop prints the problem and solution, and returns the solution:')
gtpyhop.verbose = 1
gtpyhop.pyhop(state1,[('travel','me','home','park')])

print('- If verbose=2, GTPyhop also prints a note at each recursive call:')
gtpyhop.verbose = 2
gtpyhop.pyhop(state1,[('travel','me','home','park')])

print('- If verbose=3, GTPyhop also prints the intermediate states:')
gtpyhop.verbose = 3
gtpyhop.pyhop(state1,[('travel','me','home','park')])




# SPDX-FileCopyrightText: 2021 University of Maryland
# SPDX-License-Identifier: BSD-3-Clause-Clear

# GTPyhop, version 1.1
# Author: Dana Nau <nau@umd.edu>, July 7, 2021
#
# GTPyhop as Python package
# Version: 1.2.0
# Author: Eric Jacopin, July 2025
# Main new features:
#  - Iterative planning mode
#  - Global variables helper functions
#  - Re-entrant regression tests
# New functions:
#  - print_domain_names
#  - find_domain_by_name, is_domain_created
#  - set_current_domain, get_current_domain
#  - set_recursive_planning, get_recursive_planning, reset_planning_strategy
#  - set_verbose_level, get_verbose_level
#  - seek_plan_iterative,
#  -   - _refine_multigoal_and_continue_iterative
#      - _refine_unigoal_and_continue_iterative
#      - _refine_task_and_continue_iterative
#      - _apply_action_and_continue_iterative
# Renaming:
#  -                      seek_plan >> seek_plan_recursive
#  -     _apply_action_and_continue >> apply_action_and_continue_recursive
#  - _refine_multigoal_and_continue >> refine_multigoal_and_continue_recursive
#  -   _refine_unigoal_and_continue >> refine_unigoal_and_continue_recursive
#  -      _refine_task_and_continue >> refine_task_and_continue_recursive


"""
GTPyhop is an automated planning system that can plan for both tasks and
goals. It requires Python 3. 

Accompanying this file are a README.md file giving an overview of GTPyhop,
and several examples of how to use GTPyhop. To run them, try importing any
of the modules in the Examples directory.
"""

# For use in debugging:
# from IPython import embed
# from IPython.terminal.debugger import set_trace

import copy, sys, pprint, re

################################################################################
# How much information to print while the program is running

verbose = 1


def set_verbose_level(level):
    """
    Set the verbosity (initial value is 1) level to determines how much debugging
    information GTPyhop will print:
    - level = 0: print nothing
    - level = 1: print the initial parameters and the answer
    - level = 2: also print a message on each recursive call
    - level = 3: also print some info about intermediate computations
   """
    global verbose
    if level < 0 or level > 3:
        raise ValueError("Verbose level must be between 0 and 3.")
    verbose = level
    print(f"Verbose level set to {verbose}.")


def get_verbose_level():
    """
    Returns the current verbosity level.
    """
    return verbose 

################################################################################
# States and goals

# Sequence number to use when making copies of states.
_next_state_number = 0

class State():
    """
    s = State(state_name, **kwargs) creates an object that contains the
    state-variable bindings for a state-of-the-world.
      - state_name is the name to use for the new state.
      - The keyword args are the names and initial values of state variables.
        A state-variable's initial value is usually {}, but it can also
        be a dictionary of arguments and their initial values.
    
    Example: here are three equivalent ways to specify a state named 'foo'
    in which boxes b and c are located in room2 and room3:
        First:
           s = State('foo')
           s.loc = {}   # create a dictionary for things like loc['b']
           s.loc['b'] = 'room2'
           s.loc['c'] = 'room3'
        Second:
           s = State('foo',loc={})
           s.loc['b'] = 'room2'
           s.loc['c'] = 'room3'
        Third:
           s = State('foo',loc={'b':'room2', 'c':'room3'})
    """
    
    def __init__(self, state_name, **kwargs):
        """
        state_name is the name to use for the state. The keyword
        args are the names and initial values of state variables.
        """
        self.__name__ = state_name
        vars(self).update(kwargs)
            
    def __str__(self):
        return f"<State {self.__name__}>"
        
    def __repr__(self):
        return _make_repr(self, 'State')

    def copy(self,new_name=None):
        """
        Make a copy of the state. For its name, use new_name if it is given.
        Otherwise use the old name, with a suffix '_copy#' where # is an integer.
        """
        global _next_state_number
        the_copy = copy.deepcopy(self)
        if new_name:
            the_copy.__name__ = new_name
        else:
            the_copy.__name__ = _name_for_copy(the_copy.__name__, _next_state_number)
            _next_state_number += 1
        return the_copy

    def display(self, heading=None):
        """
        Print the state's state-variables and their values.
         - heading (optional) is a heading to print beforehand.
        """
        _print_object(self, heading=heading)

    def state_vars(self):
        """Return a list of all state-variable names in the state"""
        return [v for v in vars(self) if v != '__name__']


# Sequence number to use when making copies of multigoals.
_next_multigoal_number = 0

class Multigoal():
    """
    g = Multigoal(goal_name, **kwargs) creates an object that represents
    a conjunctive goal, i.e., the goal of reaching a state that contains
    all of the state-variable bindings in g.
      - goal_name is the name to use for the new multigoal.
      - The keyword args are name and desired values of state variables.

    Example: here are three equivalent ways to specify a goal named 'goal1'
    in which boxes b and c are located in room2 and room3:
        First:
           g = Multigoal('goal1')
           g.loc = {}   # create a dictionary for things like loc['b']
           g.loc['b'] = 'room2'
           g.loc['c'] = 'room3'
        Second:
           g = Multigoal('goal1', loc={})
           g.loc['b'] = 'room2'
           g.loc['c'] = 'room3'
        Third:
           g = Multigoal('goal1',loc={'b':'room2', 'c':'room3'})
    """

    def __init__(self, multigoal_name, **kwargs):
        """
        multigoal_name is the name to use for the multigoal. The keyword
        args are the names and desired values of state variables.
        """
        self.__name__ = multigoal_name
        vars(self).update(kwargs)
            
    def __str__(self):
        return f"<Multigoal {self.__name__}>"
        
    def __repr__(self):
        return _make_repr(self, 'Multigoal')

    def copy(self,new_name=None):
        """
        Make a copy of the multigoal. For its name, use new_name if it is given.
        Otherwise use the old name, with a suffix '_copy#' where # is an integer.
        """
        global _next_multigoal_number
        the_copy = copy.deepcopy(self)
        if new_name:
            the_copy.__name__ = new_name
        else:
            the_copy.__name__ = _name_for_copy(the_copy.__name__, _next_multigoal_number)
            _next_multigoal_number += 1
        return the_copy

    def display(self, heading=None):
        """
        Print the multigoal's state-variables and their values.
         - heading (optional) is a heading to print beforehand.
        """
        _print_object(self, heading=heading)

    def state_vars(self):
        """Return a list of all state-variable names in the multigoal"""
        return [v for v in vars(self) if v != '__name__']


################################################################################
# Auxiliary functions for state and multigoal objects.


def _make_repr(object, class_name):
    """Return a string that can be used to reconstruct the object"""
    x = f"{class_name}('{object.__name__}', "
    x += ', '.join([f'{v}={vars(object)[v]}' for v in vars(object) if v != '__name__'])
    x += ')'
    return x
    

def _name_for_copy(old_name,next_integer):
    """
    Create a name to use for a copy of an object.
    - old_name is the name of the old object.
    - next_integer is the number to use at the end of the new name.
    """
    # if old_name ends in '_copy#' where # is an integer, then
    # just replace # with next_integer
    if re.findall('_copy_[0-9]*$',old_name):
        new_name = re.sub('_[0-9]*$', f'_{next_integer}', old_name)
    # otherwise use old_name with '_copy' and next_integer appended
    else:
        new_name = f'{old_name}_copy_{next_integer}'
    return new_name


def _print_object(object, heading=None):
    """
    Print the state-variables and values in 'object', which may be either a
    state or a multigoal. 'heading' is an optional heading to print beforehand.
    """
    if heading == None:
        heading = get_type(object)
    if object != False:
        title = f"{heading} {object.__name__}:"
        dashes = '-'*len(title)
        print(title)
        print(dashes)
        for (varname,val) in vars(object).items():
            if varname != '__name__':
                print(f"  - {varname} = {val}")
        print('')
    else: 
        print('{heading} = False','\n')


# print_state and print_multigoal are identical except for their names.
print_state = _print_object
print_multigoal = _print_object

def get_type(object):
    """Return object's type name"""
    return type(object).__name__


################################################################################
# A class for holding planning-and-acting domains.


class Domain():
    """
    d = Domain(domain_name) creates an object to contain the actions, commands,
    and methods for a planning-and-acting domain. 'domain_name' is the name to
    use for the new domain.
    """

    def __init__(self,domain_name):
        """domain_name is the name to use for the domain."""

        global _domains, current_domain
        
        self.__name__ = domain_name

        _domains.append(self)
        current_domain = self
        
        # dictionary that maps each action name to the corresponding function
        self._action_dict = {}    
            
        # dictionary that maps each command name to the corresponding function
        self._command_dict = {}
        
        # dictionary that maps each task name to a list of relevant methods
        # _verify_g and _verify_mg are described later in this file.
        self._task_method_dict = \
            {'_verify_g': [_m_verify_g], '_verify_mg': [_m_verify_mg]}
        
        # dictionary that maps each unigoal name to a list of relevant methods
        self._unigoal_method_dict = {}
        
        # list of all methods for multigoals
        self._multigoal_method_list = []

    def __str__(self):
        return f"<Domain {self.__name__}>"
        
    def __repr__(self):
        return _make_repr(self, 'Domain')

    def copy(self,new_name=None):
        """
        Make a copy of the domain. For its name, use new_name if it is given.
        Otherwise use the old name, with a suffix '_copy#' where # is an integer.
        """
        global _next_domain_number
        the_copy = copy.deepcopy(self)
        if new_name:
            the_copy.__name__ = new_name
        else:
            the_copy.__name__ = _name_for_copy(the_copy.__name__, _next_domain_number)
            _next_domain_number += 1
        return the_copy

    def display(self):
        """Print the domain's actions, commands, and methods."""
        print_domain(self)
        

# Sequence number to use when making copies of domains.
_next_domain_number = 0

# A list of all domains that have been created
_domains = []


def print_domain_names():
    """
    Print the names of all domains that have been created.
    """
    if _domains:
        print('-- Domain names:', ', '.join([d.__name__ for d in _domains]))
    else:
        print('-- There are no domains --')


def find_domain_by_name(domain_name):
    """
    Search for a domain by its name in the _domains list.

    :param domain_name: The name of the domain to search for.
    :return: The domain instance if found, otherwise None.
    """
    for domain in _domains:
        if domain.__name__ == domain_name:
            return domain
    return None

def is_domain_created(domain_name):
    """
    Check if a domain with the given name has been created in the _domains list.

    :param domain_name: The name of the domain to check.
    :return: True if the domain is found, otherwise False.
    """
    for domain in _domains:
        if domain.__name__ == domain_name:
            return True
    return False

current_domain = None
"""
The Domain object that find_plan, run_lazy_lookahead, etc., will use.
"""

def set_current_domain(domain):
    """
    Set the current domain to the specified domain object.
    
    :param domain: The Domain object to set as the current domain.
    """
    global current_domain
    if not isinstance(domain, Domain):
        raise TypeError("The provided argument is not a Domain instance.")
    current_domain = domain

def get_current_domain():
    """
    Get the current domain object.
    
    :return: The current Domain object.
    """
    return current_domain

################################################################################
# Functions to print information about a domain


def print_domain(domain=None):
    """
    Print domain's actions, commands, and methods. The optional 'domain'
    argument defaults to the current domain
    """
    if domain == None:
        domain = current_domain
    print(f'\nDomain name: {domain.__name__}')
    print_actions(domain)
    print_commands(domain)
    print_methods(domain)

def print_actions(domain=None):
    """Print the names of all the actions"""
    if domain == None:
        domain = current_domain
    if domain._action_dict:
        print('-- Actions:', ', '.join(domain._action_dict))
    else:
        print('-- There are no actions --')

def print_operators():
    if verbose > 0:
        print("""
        >> print_operators exists to provide backward compatibility
        >> with Pyhop. In the future, please use print_actions instead.""")
    return print_actions()

def print_commands(domain=None):
    """Print the names of all the commands"""
    if domain == None:
        domain = current_domain
    if domain._command_dict:
        print('-- Commands:', ', '.join(domain._command_dict))
    else:
        print('-- There are no commands --')

def _print_task_methods(domain):
    """Print a table of the task_methods for each task"""
    if domain._task_method_dict:
        print('')
        print('Task name:         Relevant task methods:')
        print('---------------    ----------------------')
        for task in domain._task_method_dict:
            print(f'{task:<19}' + ', '.join(    \
                [f.__name__ for f in domain._task_method_dict[task]]))
        print('')
    else:
        print('-- There are no task methods --')

def _print_unigoal_methods(domain):
    """Print a table of the unigoal_methods for each state_variable_name"""
    if domain._unigoal_method_dict:
        print('State var name:    Relevant unigoal methods:')
        print('---------------    -------------------------')
        for var in domain._unigoal_method_dict:
            print(f'{var:<19}' + ', '.join( \
                [f.__name__ for f in domain._unigoal_method_dict[var]]))
        print('')
    else:
        print('-- There are no unigoal methods --')

def _print_multigoal_methods(domain):
    """Print the names of all the multigoal_methods"""
    if domain._multigoal_method_list:
        print('-- Multigoal methods:', ', '.join(  \
                [f.__name__ for f in domain._multigoal_method_list]))
    else:
        print('-- There are no multigoal methods --')
    
def print_methods(domain=None):
    """Print tables showing what all the methods are"""
    if domain == None:
        domain = current_domain
    _print_task_methods(domain)
    _print_unigoal_methods(domain)
    _print_multigoal_methods(domain)


################################################################################
# Functions to declare actions, commands, tasks, unigoals, multigoals


def declare_actions(*actions):
    """
    declare_actions adds each member of 'actions' to the current domain's list
    of actions. For example, this says that pickup and putdown are actions:
        declare_actions(pickup,putdown)
        
    declare_actions can be called multiple times to add more actions.
    
    You can see the current domain's list of actions by executing
        current_domain.display()
    """
    if current_domain == None:
        raise Exception(f"cannot declare actions until a domain has been created.")
    current_domain._action_dict.update({act.__name__:act for act in actions})
    return current_domain._action_dict



def declare_operators(*actions):
    if verbose > 0:
        print("""
        >> declare_operators exists to provide backward compatibility
        >> with Pyhop. In the future, please use declare_actions instead.""")
    return declare_actions(*actions)


def declare_commands(*commands):
    """
    declare_commands adds each member of 'commands' to the current domain's
    list of commands.  Each member of 'commands' should be a function whose
    name has the form c_foo, where foo is the name of an action. For example,
    this says that c_pickup and c_putdown are commands:
        declare_commands(c_pickup,c_putdown)
    
    declare_commands can be called several times to add more commands.

    You can see the current domain's list of commands by executing
        current_domain.display()

    """
    if current_domain == None:
        raise Exception(f"cannot declare commands until a domain has been created.")
    current_domain._command_dict.update({cmd.__name__:cmd for cmd in commands})
    return current_domain._command_dict


def declare_task_methods(task_name, *methods):
    """
    'task_name' should be a character string, and 'methods' should be a list
    of functions. declare_task_methods adds each member of 'methods' to the
    current domain's list of methods to use for tasks of the form
        (task_name, arg1, ..., argn).     

    Example:
        declare_task_methods('travel', travel_by_car, travel_by_foot)
    says that travel_by_car and travel_by_foot are methods and that GTPyhop
    should try using them for any task whose task name is 'travel', e.g.,
        ('travel', 'alice', 'store')
        ('travel', 'alice', 'umd', 'ucla')
        ('travel', 'alice', 'umd', 'ucla', 'slowly')
        ('travel', 'bob', 'home', 'park', 'looking', 'at', 'birds')

    This is like Pyhop's declare_methods function, except that it can be
    called several times to declare more methods for the same task.
    """
    if current_domain == None:
        raise Exception(f"cannot declare methods until a domain has been created.")
    if task_name in current_domain._task_method_dict:
        old_methods = current_domain._task_method_dict[task_name]
        # even though current_domain._task_method_dict[task_name] is a list,
        # we don't want to add any methods that are already in it
        new_methods = [m for m in methods if m not in old_methods]
        current_domain._task_method_dict[task_name].extend(new_methods)
    else:
        current_domain._task_method_dict.update({task_name:list(methods)})
    return current_domain._task_method_dict


def declare_methods(task, *methods):
    if verbose > 0:
        print("""
        >> declare_methods exists to provide backward compatibility with
        >> Pyhop. In the future, please use declare_task_methods instead.""")
    return declare_task_methods(task, *methods)


def declare_unigoal_methods(state_var_name, *methods):
    """
    'state_var_name' should be a character string, and 'methods' should be a
    list of functions. declare_unigoal_method adds each member of 'methods'
    to the current domain's list of relevant methods for goals of the form
        (state_var_name, arg, value)
    where 'arg' and 'value' are the state variable's argument and the desired
    value. For example,
        declare_unigoal_method('loc',travel_by_car)
    says that travel_by_car is relevant for goals such as these:
        ('loc', 'alice', 'ucla')
        ('loc', 'bob', 'home')

    The above kind of goal, i.e., a desired value for a single state
    variable, is called a "unigoal". To achieve a unigoal, GTPyhop will go
    through the unigoal's list of relevant methods one by one, trying each
    method until it finds one that is successful.

    To see each unigoal's list of relevant methods, use
        current_domain.display()    
    """
    if current_domain == None:
        raise Exception(f"cannot declare methods until a domain has been created.")
    if state_var_name not in current_domain._unigoal_method_dict:
        current_domain._unigoal_method_dict.update({state_var_name:list(methods)})
    else:
        old_methods = current_domain._unigoal_method_dict[state_var_name]
        new_methods = [m for m in methods if m not in old_methods]
        current_domain._unigoal_method_dict[state_var_name].extend(new_methods)
    return current_domain._unigoal_method_dict    


def declare_multigoal_methods(*methods):
    """
    declare_multigoal_methods adds each method in 'methods' to the current
    domain's list of multigoal methods. For example, this says that
    stack_all_blocks and unstack_all_blocks are multigoal methods:
        declare_multigoal_methods(stack_all_blocks, unstack_all_blocks)
    
    When GTPyhop tries to achieve a multigoal, it will go through the list
    of multigoal methods one by one, trying each method until it finds one
    that is successful. You can see the list by executing
        current_domain.display()

    declare_multigoal_methods can be called multiple times to add more
    multigoal methods to the list.
    
    For more information, see the docstring for the Multigoal class.
    """
    if current_domain == None:
        raise Exception(    \
                f"cannot declare methods until a domain has been created.")
    new_mg_methods = [m for m in methods if m not in \
                      current_domain._multigoal_method_list]
    current_domain._multigoal_method_list.extend(new_mg_methods)
    return current_domain._multigoal_method_list    

    
################################################################################
# A built-in multigoal method and its helper function.


def m_split_multigoal(state,multigoal):
    """
    m_split_multigoal is the only multigoal method that GTPyhop provides,
    and GTPyhop won't use it unless the user declares it explicitly using
        declare_multigoal_methods(m_split_multigoal)

    The method's purpose is to try to achieve a multigoal by achieving each
    of the multigoal's individual goals sequentially. Parameters:
        - 'state' is the current state
        - 'multigoal' is the multigoal to achieve 

    If multigoal is true in the current state, m_split_multigoal returns
    []. Otherwise, it returns a goal list
        [g_1, ..., g_n, multigoal],

    where g_1, ..., g_n are all of the goals in multigoal that aren't true
    in the current state. This tells the planner to achieve g_1, ..., g_n
    sequentially, then try to achieve multigoal again. Usually this means
    m_split_multigal will be used repeatedly, until it succeeds in producing
    a state in which all of the goals in multigoal are simultaneously true.

    The main problem with m_split_multigoal is that it isn't smart about
    choosing the order in which to achieve g_1, ..., g_n. Some orderings may
    work much better than others. Thus, rather than using the method as it's
    defined below, one might want to modify it to choose a good order, e.g.,
    by using domain-specific information or a heuristic function.
    """
    goal_dict = _goals_not_achieved(state,multigoal)
    goal_list = []
    for state_var_name in goal_dict:
        for arg in goal_dict[state_var_name]:
            val = goal_dict[state_var_name][arg]
            goal_list.append((state_var_name,arg,val))
    if goal_list:
        # achieve goals, then check whether they're all simultaneously true
        return goal_list + [multigoal]
    return goal_list


# helper function for m_split_multigoal above:

def _goals_not_achieved(state,multigoal):
    """
    _goals_not_achieved takes two arguments: a state s and a multigoal g.
    It returns a dictionary of the goals in g that aren't true in s.
    For example, suppose
        s.loc['c0'] = 'room0', g.loc['c0'] = 'room0',
        s.loc['c1'] = 'room1', g.loc['c1'] = 'room3',
        s.loc['c2'] = 'room2', g.loc['c2'] = 'room4'.
    Then _goals_not_achieved(s, g) will return
        {'loc': {'c1': 'room3', 'c2': 'room4'}}    
    """
    unachieved = {}
    for name in vars(multigoal):
        if name != '__name__':
            for arg in vars(multigoal).get(name):
                val = vars(multigoal).get(name).get(arg)
                if val != vars(state).get(name).get(arg):
                    # want arg_value_pairs.name[arg] = val
                    if not unachieved.get(name):
                        unachieved.update({name:{}})
                    unachieved.get(name).update({arg:val})
    return unachieved


################################################################################
# Functions to verify whether unigoal_methods achieve the goals they are
# supposed to achieve.


verify_goals = True
"""
If verify_goals is True, then whenever the planner uses a method m to refine
a unigoal or multigoal, it will insert a "verification" task into the
current partial plan. If verify_goals is False, the planner won't insert any
verification tasks into the plan.

The purpose of the verification task is to raise an exception if the
refinement produced by m doesn't achieve the goal or multigoal that it is
supposed to achieve. The verification task won't insert anything into the
final plan; it just will verify whether m did what it was supposed to do.
"""


def _m_verify_g(state, method, state_var, arg, desired_val, depth):
    """
    _m_verify_g is a method that GTPyhop uses to check whether a
    unigoal method has achieved the goal for which it was used.
    """
    if vars(state)[state_var][arg] != desired_val:
        raise Exception(f"depth {depth}: method {method} didn't achieve",
                f"goal {state_var}[{arg}] = {desired_val}")
    if verbose >= 3:
        print(f"depth {depth}: method {method} achieved",
                f"goal {state_var}[{arg}] = {desired_val}")
    return []       # i.e., don't create any subtasks or subgoals


def _m_verify_mg(state, method, multigoal, depth):
    """
    _m_verify_g is a method that GTPyhop uses to check whether a multigoal
    method has achieved the multigoal for which it was used.
    """
    goal_dict = _goals_not_achieved(state,multigoal)
    if goal_dict:
        raise Exception(f"depth {depth}: method {method} " + \
                        f"didn't achieve {multigoal}]")
    if verbose >= 3:
        print(f"depth {depth}: method {method} achieved {multigoal}")
    return []


################################################################################
# Applying actions, commands, and methods recursively to seek a plan


def _apply_action_and_continue_recursive(state, task1, todo_list, plan, depth):
    """
    _apply_action_and_continue is called only when task1's name matches an
    action name. It applies the action by retrieving the action's function
    definition and calling it on the arguments, then calls seek_plan
    recursively on todo_list.
    """
    if verbose >= 3:
        print(f'depth {depth} action {task1}: ', end='')
    action = current_domain._action_dict[task1[0]]
    newstate = action(state.copy(),*task1[1:])
    if newstate:
        if verbose >= 3:
            print('applied')
            newstate.display()
        return seek_plan_recursive(newstate, todo_list, plan+[task1], depth+1)
    if verbose >= 3:
        print('not applicable')
    return False


def _refine_task_and_continue_recursive(state, task1, todo_list, plan, depth):
    """
    If task1 is in the task-method dictionary, then iterate through the list
    of relevant methods to find one that's applicable, apply it to get
    additional todo_list items, and call seek_plan recursively on
            [the additional items] + todo_list.

    If the call to seek_plan_recursive fails, go on to the next method in the list.
    """
    relevant = current_domain._task_method_dict[task1[0]]
    if verbose >= 3:
        print(f'depth {depth} task {task1} methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3: 
            print(f'depth {depth} trying {method.__name__}: ', end='')
        subtasks = method(state, *task1[1:])
        # Can't just say "if subtasks:", because that's wrong if subtasks == []
        if subtasks != False and subtasks != None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subtasks: {subtasks}')
            result = seek_plan_recursive(state, subtasks+todo_list, plan, depth+1)
            if result != False and result != None:
                return result
        else:
            if verbose >= 3:
                print(f'not applicable')
    if verbose >= 3:
        print(f'depth {depth} could not accomplish task {task1}')        
    return False


def _refine_unigoal_and_continue_recursive(state, goal1, todo_list, plan, depth):
    """
    If goal1 is in the unigoal-method dictionary, then iterate through the
    list of relevant methods to find one that's applicable, apply it to get
    additional todo_list items, and call seek_plan recursively on
          [the additional items] + [verify_g] + todo_list,

    where [verify_g] verifies whether the method actually achieved goal1.
    If the call to seek_plan_recursive fails, go on to the next method in the list.
    """
    if verbose >= 3:
        print(f'depth {depth} goal {goal1}: ', end='')
    (state_var_name, arg, val) = goal1
    if vars(state).get(state_var_name).get(arg) == val:
        if verbose >= 3:
            print(f'already achieved')
        return seek_plan_recursive(state, todo_list, plan, depth+1)
    relevant = current_domain._unigoal_method_dict[state_var_name]
    if verbose >= 3:
        print(f'methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3: 
            print(f'depth {depth} trying method {method.__name__}: ', end='')
        subgoals = method(state,arg,val)
        # Can't just say "if subgoals:", because that's wrong if subgoals == []
        if subgoals != False and subgoals != None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subgoals: {subgoals}')
            if verify_goals:
                verification = [('_verify_g', method.__name__, \
                                 state_var_name, arg, val, depth)]
            else:
                verification = []
            todo_list = subgoals + verification + todo_list
            result = seek_plan_recursive(state, todo_list, plan, depth+1)
            if result != False and result != None:
                return result
        else:
            if verbose >= 3:
                print(f'not applicable')        
    if verbose >= 3:
        print(f'depth {depth} could not achieve goal {goal1}')        
    return False


def _refine_multigoal_and_continue_recursive(state, goal1, todo_list, plan, depth):
    """
    If goal1 is a multigoal, then iterate through the list of multigoal
    methods to find one that's applicable, apply it to get additional
    todo_list items, and call seek_plan recursively on
          [the additional items] + [verify_mg] + todo_list,

    where [verify_mg] verifies whether the method actually achieved goal1.
    If the call to seek_plan_recursive fails, go on to the next method in the list.
    """
    if verbose >= 3:
        print(f'depth {depth} multigoal {goal1}: ', end='')
    relevant = current_domain._multigoal_method_list
    if verbose >= 3:
        print(f'methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3: 
            print(f'depth {depth} trying method {method.__name__}: ', end='')
        subgoals = method(state,goal1)
        # Can't just say "if subgoals:", because that's wrong if subgoals == []
        if subgoals != False and subgoals != None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subgoals: {subgoals}')
            if verify_goals:
                verification = [('_verify_mg', method.__name__, goal1, depth)]
            else:
                verification = []
            todo_list = subgoals + verification + todo_list
            result = seek_plan_recursive(state, todo_list, plan, depth+1)
            if result != False and result != None:
                return result
        else:
            if verbose >= 3:
                print(f'not applicable')
    if verbose >= 3:
        print(f'depth {depth} could not achieve multigoal {goal1}')        
    return False

def seek_plan_recursive(state, todo_list, plan, depth):
    """
    Recursive workhorse for find_plan. Arguments:
     - state is the current state
     - todo_list is the current list of goals, tasks, and actions
     - plan is the current partial plan
     - depth is the recursion depth, for use in debugging
    """
    if verbose >= 2: 
        todo_string = '[' + ', '.join([_item_to_string(x) for x in todo_list]) + ']'
        print(f'depth {depth} todo_list ' + todo_string)
    if todo_list == []:
        if verbose >= 3:
            print(f'depth {depth} no more tasks or goals, return plan')
        return plan
    item1 = todo_list[0]
    ttype = get_type(item1)
    if ttype in {'Multigoal'}:
        return _refine_multigoal_and_continue_recursive(state, item1, todo_list[1:], plan, depth)
    elif ttype in {'list','tuple'}:
        if item1[0] in current_domain._action_dict:
            return _apply_action_and_continue_recursive(state, item1, todo_list[1:], plan, depth)
        elif item1[0] in current_domain._task_method_dict:
            return _refine_task_and_continue_recursive(state, item1, todo_list[1:], plan, depth)
        elif item1[0] in current_domain._unigoal_method_dict:
            return _refine_unigoal_and_continue_recursive(state, item1, todo_list[1:], plan, depth)
    raise Exception(    \
        f"depth {depth}: {item1} isn't an action, task, unigoal, or multigoal\n")
    return False


###############################################################################
# Applying actions, commands, and methods iteratively to seek a plan

def _apply_action_and_continue_iterative(state, task1, todo_list, plan, depth):
    if verbose >= 3:
        print(f'depth {depth} action {task1}: ', end='')
    action = current_domain._action_dict[task1[0]]
    newstate = action(state.copy(), *task1[1:])
    if newstate:
        if verbose >= 3:
            print('applied')
            newstate.display()
        return (newstate, todo_list, plan + [task1], depth + 1)
    if verbose >= 3:
        print('not applicable')
    return None

def _refine_task_and_continue_iterative(state, task1, todo_list, plan, depth):
    relevant = current_domain._task_method_dict[task1[0]]
    if verbose >= 3:
        print(f'depth {depth} task {task1} methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3:
            print(f'depth {depth} trying {method.__name__}: ', end='')
        subtasks = method(state, *task1[1:])
        if subtasks is not False and subtasks is not None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subtasks: {subtasks}')
            result = (state, subtasks + todo_list, plan, depth + 1)
            return result  # Return new state to be added to the stack
        else:
            if verbose >= 3:
                print(f'not applicable')
    if verbose >= 3:
        print(f'depth {depth} could not accomplish task {task1}')
    return None

def _refine_unigoal_and_continue_iterative(state, goal1, todo_list, plan, depth):
    if verbose >= 3:
        print(f'depth {depth} goal {goal1}: ', end='')
    (state_var_name, arg, val) = goal1
    if vars(state).get(state_var_name).get(arg) == val:
        if verbose >= 3:
            print(f'already achieved')
        return (state, todo_list, plan, depth + 1)
    relevant = current_domain._unigoal_method_dict[state_var_name]
    if verbose >= 3:
        print(f'methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3:
            print(f'depth {depth} trying method {method.__name__}: ', end='')
        subgoals = method(state, arg, val)
        if subgoals is not False and subgoals is not None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subgoals: {subgoals}')
            if verify_goals:
                verification = [('_verify_g', method.__name__, state_var_name, arg, val, depth)]
            else:
                verification = []
            new_todo_list = subgoals + verification + todo_list
            return (state, new_todo_list, plan, depth + 1)
        else:
            if verbose >= 3:
                print(f'not applicable')
    if verbose >= 3:
        print(f'depth {depth} could not achieve goal {goal1}')
    return None

def _refine_multigoal_and_continue_iterative(state, goal1, todo_list, plan, depth):
    if verbose >= 3:
        print(f'depth {depth} multigoal {goal1}: ', end='')
    relevant = current_domain._multigoal_method_list
    if verbose >= 3:
        print(f'methods {[m.__name__ for m in relevant]}')
    for method in relevant:
        if verbose >= 3:
            print(f'depth {depth} trying method {method.__name__}: ', end='')
        subgoals = method(state, goal1)
        if subgoals is not False and subgoals is not None:
            if verbose >= 3:
                print('applicable')
                print(f'depth {depth} subgoals: {subgoals}')
            if verify_goals:
                verification = [('_verify_mg', method.__name__, goal1, depth)]
            else:
                verification = []
            new_todo_list = subgoals + verification + todo_list
            return (state, new_todo_list, plan, depth + 1)
        else:
            if verbose >= 3:
                print(f'not applicable')
    if verbose >= 3:
        print(f'depth {depth} could not achieve multigoal {goal1}')
    return None

def seek_plan_iterative(initial_state, initial_todo_list, initial_plan, initial_depth):
    """
    Iterative workhorse for find_plan. Arguments:
     - state is the current state
     - todo_list is the current list of goals, tasks, and actions
     - plan is the current partial plan
     - depth is the recursion depth, for use in debugging
    """
    stack = [(initial_state, initial_todo_list, initial_plan, initial_depth)]  # (state, todo_list, plan, depth)

    while stack:
        state, todo_list, plan, depth = stack.pop()

        if verbose >= 2:
            todo_string = '[' + ', '.join([_item_to_string(x) for x in todo_list]) + ']'
            print(f'depth {depth} todo_list ' + todo_string)

        if not todo_list:
            if verbose >= 3:
                print(f'depth {depth} no more tasks or goals, return plan')
            return plan

        item1 = todo_list[0]
        ttype = get_type(item1)

        if ttype in {'Multigoal'}:
            result = _refine_multigoal_and_continue_iterative(state, item1, todo_list[1:], plan, depth)
            if result is not None:
                stack.append(result)  # Add new state to the stack
        elif ttype in {'list', 'tuple'}:
            if item1[0] in current_domain._action_dict:
                result = _apply_action_and_continue_iterative(state, item1, todo_list[1:], plan, depth)
                if result is not None:
                    stack.append(result)  # Add new state to the stack
            elif item1[0] in current_domain._task_method_dict:
                result = _refine_task_and_continue_iterative(state, item1, todo_list[1:], plan, depth)
                if result is not None:
                    stack.append(result)  # Add new state to the stack
            elif item1[0] in current_domain._unigoal_method_dict:
                result = _refine_unigoal_and_continue_iterative(state, item1, todo_list[1:], plan, depth)
                if result is not None:
                    stack.append(result)  # Add new state to the stack

    return False

############################################################
# The planning system

_current_seek_plan = None


def set_recursive_planning(use_recursive):
    global _current_seek_plan
    if use_recursive:
        _current_seek_plan = seek_plan_recursive
        print("Using recursive seek_plan.")
    else:
        _current_seek_plan = seek_plan_iterative
        print("Using iterative seek_plan.")


def get_recursive_planning():
    """
    Returns True if the current seek_plan is recursive, False if it is iterative.
    """
    if None == _current_seek_plan:
        raise Exception("No planning strategy (iterative or else recursive) has been set. Use set_recursive_planning(True|False) to set it.")
    return _current_seek_plan == seek_plan_recursive


def reset_planning_strategy():
    """
    Resets the planning strategy to None, so that the user must set it again
    using set_recursive_planning(True|False).
    """
    global _current_seek_plan
    _current_seek_plan = None


def seek_plan(state, todolist, plan, depth):
    return _current_seek_plan(state, todolist, plan, depth)


def find_plan(state, todo_list):
    """
    find_plan tries to find a plan that accomplishes the items in todo_list,
    starting from the given state, using whatever methods and actions you
    declared previously. If successful, it returns the plan. Otherwise it
    returns False. Arguments:
     - 'state' is a state;
     - 'todo_list' is a list of goals, tasks, and actions.
    """
    if None == _current_seek_plan:
        raise Exception("No planning strategy (iterative or else recursive) has been set. Use set_recursive_planning(True|False) to set it.")
    if verbose >= 1:
        todo_string = '[' + ', '.join([_item_to_string(x) for x in todo_list]) + ']'
        print(f'FP> find_plan, verbose={verbose}:')
        print(f'    state = {state.__name__}\n    todo_list = {todo_string}')
    result = _current_seek_plan(state, todo_list, [], 0)
    if verbose >= 1: print('FP> result =',result,'\n')
    return result


def pyhop(state, todo_list):
    if verbose > 0:
        print("""
        >> The function 'pyhop' exists to provide backward compatibility
        >> with Pyhop. In the future, please use find_plan instead.""")
    return find_plan(state, todo_list)


def _item_to_string(item):
    """Return a string representation of a task or goal."""
    ttype = get_type(item)
    if ttype == 'list':
        return str([str(x) for x in item])
    elif ttype == 'tuple':
        return str(tuple([str(x) for x in item]))
    else:       # a multigoal
        return str(item)

################################################################################
# An actor


def run_lazy_lookahead(state, todo_list, max_tries=10):
    """
    An adaptation of the run_lazy_lookahead algorithm from Ghallab et al.
    (2016), Automated Planning and Acting. It works roughly like this:
        loop:
            plan = find_plan(state, todo_list)
            if plan = [] then return state    // the new current state 
            for each action in plan:
                try to execute the corresponding command
                if the command fails, continue the outer loop
    Arguments: 
      - 'state' is a state;
      - 'todo_list' is a list of tasks, goals, and multigoals;
      - max_tries is a bound on how many times to execute the outer loop.
      
    Note: whenever run_lazy_lookahead encounters an action for which there is
    no corresponding command definition, it uses the action definition instead.
    """
    if verbose >= 1: 
        print(f"RLL> run_lazy_lookahead, verbose = {verbose}, max_tries = {max_tries}")
        print(f"RLL> initial state: {state.__name__}")
        print('RLL> To do:', todo_list)

    for tries in range(1,max_tries+1):
        if verbose >= 1: 
            ordinals = {1:'st',2:'nd',3:'rd'}
            if ordinals.get(tries):
                print(f"RLL> {tries}{ordinals.get(tries)} call to find_plan:\n")
            else:
                print(f"RLL> {tries}th call to find_plan:\n")
        plan = find_plan(state, todo_list)
        if plan == False or plan == None:
            if verbose >= 1:
                raise Exception(
                        f"run_lazy_lookahead: find_plan has failed")
            return state
        if plan == []:
            if verbose >= 1: 
                print(f'RLL> Empty plan => success',
                      f'after {tries} calls to find_plan.')
            if verbose >= 2: state.display(heading='> final state')
            return state
        for action in plan:
            command_name = 'c_' + action[0]
            command_func = current_domain._command_dict.get(command_name)
            if command_func == None:
                if verbose >= 1: 
                    print(f'RLL> {command_name} not defined, using {action[0]} instead\n')
                command_func = current_domain._action_dict.get(action[0])
                
            if verbose >= 1:
                print('RLL> Command:', [command_name] + list(action[1:]))
            new_state = _apply_command_and_continue_rll(state, command_func, action[1:])
            if new_state == False:
                if verbose >= 1: 
                    print(f'RLL> WARNING: command {command_name} failed; will call find_plan.')
                    break
            else:
                if verbose >= 2: 
                    new_state.display()
                state = new_state
        # if state != False then we're here because the plan ended
        if verbose >= 1 and state:
            print(f'RLL> Plan ended; will call find_plan again.')
        
    if verbose >= 1: print('RLL> Too many tries, giving up.')
    if verbose >= 2: state.display(heading='RLL> final state')
    return state


def _apply_command_and_continue_rll(state, command, args):
    """
    _apply_command_and_continue applies 'command' by retrieving its
    function definition and calling it on the arguments.
    """
    if verbose >= 3:
        print(f"_apply_command_and_continue {command.__name__}, args = {args}")
    next_state = command(state.copy(),*args)
    if next_state:
        if verbose >= 3:
            print('applied')
            next_state.display()
        return next_state
    else:
        if verbose >= 3:
            print('not applicable')
        return False


###############################################################################
# Print brief information about how to interpret the program's output

print(f"\nImported GTPyhop version 1.2.0")
print(f"Messages from find_plan will be prefaced with 'FP>'.")
print(f"Messages from run_lazy_lookahead will be prefaced with 'RLL>'.")
set_recursive_planning(False) # default is to use iterative planning


__version__ = "1.2.0rc1"

from .main import (
    verbose,
    set_verbose_level,
    get_verbose_level,
    Domain,
    current_domain,
    set_current_domain,
    get_current_domain,
    print_domain,
    print_domain_names,
    find_domain_by_name,
    is_domain_created,
    print_actions,
    print_operators,
    print_commands,
    print_methods,
    State,
    print_state,
    Multigoal,
    print_multigoal,
    get_type,
    declare_actions,
    declare_operators,
    declare_commands,
    declare_task_methods,
    declare_methods,
    declare_unigoal_methods,
    declare_multigoal_methods,
    m_split_multigoal,
    set_recursive_planning,
    get_recursive_planning,
    reset_planning_strategy,
    find_plan,
    pyhop,
    run_lazy_lookahead
)


"""
Examples file for blocks_gtn.
-- Dana Nau <nau@umd.edu>, July 20, 2021
"""

# Uncomment this to use it in debugging:
# from IPython import embed
# from IPython.terminal.debugger import set_trace

import gtpyhop
import gtpyhop.test_harness as th   # code for use in paging and debugging

print(">>> __name__ = ", __name__)  # Debugging line

# We must declare the current domain before importing methods and actions.
# To make the code more portable, we don't hard-code the domain name, but
# instead use the name of the package.
the_domain = gtpyhop.Domain(__package__)
print(">>> the_domain = ", the_domain)  # Debugging line
print(">>> the_domain.__name__ = ", the_domain.__name__)  # Debugging line
from .methods import *
from .actions import *

print('-----------------------------------------------------------------------')
print(f"Created '{gtpyhop.current_domain}'. To run the examples, type this:")
print(f'{the_domain.__name__}.main()')


#############     beginning of tests     ################

def main(do_pauses=True):
    """
    Run various examples.
    main() will pause occasionally to let you examine the output.
    main(False) will run straight through to the end, without stopping.
    """

    # If we've changed to some other domain, this will change us back.
    print(f"Changing current domain to {the_domain}, if it isn't that already.")

    gtpyhop.set_current_domain(gtpyhop.find_domain_by_name(gtpyhop.Domain(__package__).__name__))
    gtpyhop.print_domain()

    print("\nLet's call find_plan on some simple things that should fail.\n")

    state1 = gtpyhop.State('state1')
    state1.pos={'a':'b', 'b':'table', 'c':'table'}
    state1.clear={'c':True, 'b':False,'a':True}
    state1.holding={'hand':False}

    state1.display('Initial state is')

    plan = gtpyhop.find_plan(state1,[('pickup','a')])
    th.check_result(plan,False)

    plan = gtpyhop.find_plan(state1,[('pickup','b')])
    th.check_result(plan,False)

    plan = gtpyhop.find_plan(state1,[('take','b')])
    th.check_result(plan,False)

    th.pause(do_pauses)
    print("""
Next, some simple things that should succeed. As a reminder, in state1,
block a is on block b, block b is on the table, and block c is on the table.
""")
    
    plan = gtpyhop.find_plan(state1,[('pickup','c')])
    th.check_result(plan, [('pickup','c')])

    plan = gtpyhop.find_plan(state1,[('take','a')])
    th.check_result(plan, [('unstack','a', 'b')])

    plan = gtpyhop.find_plan(state1,[('take','c')])
    th.check_result(plan, [('pickup','c')])

    plan = gtpyhop.find_plan(state1,[('take','a'),('put','a','table')])
    th.check_result(plan, [('unstack','a', 'b'), ('putdown','a')])
    th.pause(do_pauses)


    print("""
A Multigoal is a data structure that specifies desired values for some of
the state variables. Below, goal1a says we want the blocks in the
configuration "c on b, b on a, a on the table", and goal1b says we want "c
on b, b on a" without specifying where block a should be. However, goal1a
and goal1b have the same solution plans, because "c on b, b on a" entails "a
on the table".
""")

    state1.display("Initial state is")

    goal1a = gtpyhop.Multigoal('goal1a')
    goal1a.pos={'c':'b', 'b':'a', 'a':'table'}

    goal1a.display()
    
    goal1b = gtpyhop.Multigoal('goal1b')
    goal1b.pos={'c':'b', 'b':'a'}

    goal1b.display()

    # Tell the test harness what answer to expect, so it can signal an error
    # if gtpyhop returns an incorrect answer. Checks like this have been very
    # very helpful for debugging both gtpyhop and the various example domains.

    expected = [('unstack', 'a', 'b'), ('putdown', 'a'), ('pickup', 'b'), ('stack', 'b', 'a'), ('pickup', 'c'), ('stack', 'c', 'b')] 

    plan1 = gtpyhop.find_plan(state1,[goal1a])
    th.check_result(plan1,expected)

    plan2 = gtpyhop.find_plan(state1,[goal1b])
    th.check_result(plan2,expected)
    th.pause(do_pauses)


    print("""
Run find_plan on the famous Sussman anomaly.
""")

    sus_s0 = gtpyhop.State('Sussman anomaly initial state')
    sus_s0.pos={'c':'a', 'a':'table', 'b':'table'}
    sus_s0.clear={'c':True, 'a':False,'b':True}
    sus_s0.holding={'hand':False}

    sus_s0.display()
    
    sus_sg = gtpyhop.Multigoal('Sussman anomaly multigoal')
    sus_sg.pos={'a':'b', 'b':'c'}

    sus_sg.display()

    expected = [('unstack', 'c', 'a'), ('putdown', 'c'), ('pickup', 'b'), ('stack', 'b', 'c'), ('pickup', 'a'), ('stack', 'a', 'b')] 

    sussman_plan = gtpyhop.find_plan(sus_s0,[sus_sg])
    th.check_result(sussman_plan,expected)

    th.pause(do_pauses)

    print("""
Run find_plan on two more planning problems. Like before, goal2b omits some
of the conditions in goal2a, but both goals should produce the same plan.
""")

    state2 = gtpyhop.State('state2')
    state2.pos={'a':'c', 'b':'d', 'c':'table', 'd':'table'}
    state2.clear={'a':True, 'c':False,'b':True, 'd':False}
    state2.holding={'hand':False}

    state2.display('Initial state is')
    
    goal2a = gtpyhop.Multigoal('goal2a')
    goal2a.pos={'b':'c', 'a':'d', 'c':'table', 'd':'table'}
    goal2a.clear={'a':True, 'c':False,'b':True, 'd':False}
    goal2a.holding={'hand':False}

    goal2a.display()
    
    goal2b = gtpyhop.Multigoal('goal2b')
    goal2b.pos={'b':'c', 'a':'d'}

    goal2b.display()
    
    ### goal2b omits some of the conditions of goal2a,
    ### but those conditions will need to be achieved anyway.

    expected = [('unstack', 'a', 'c'), ('putdown', 'a'), ('unstack', 'b', 'd'), ('stack', 'b', 'c'), ('pickup', 'a'), ('stack', 'a', 'd')] 

    plan1 = gtpyhop.find_plan(state2,[goal2a])
    th.check_result(plan1,expected)

    plan2 = gtpyhop.find_plan(state2,[goal2b])
    th.check_result(plan2,expected)
    th.pause(do_pauses)


    print("\nRun find_plan on problem bw_large_d from the SHOP distribution:\n")

    state3 = gtpyhop.State('state3')
    state3.pos = {1:12, 12:13, 13:'table', 11:10, 10:5, 5:4, 4:14, 14:15, 15:'table', 9:8, 8:7, 7:6, 6:'table', 19:18, 18:17, 17:16, 16:3, 3:2, 2:'table'}
    state3.clear = {x:False for x in range(1,20)}
    state3.clear.update({1:True, 11:True, 9:True, 19:True})
    state3.holding={'hand':False}

    state3.display('Initial state is')
    
    goal3 = gtpyhop.Multigoal('goal3')
    goal3.pos = {15:13, 13:8, 8:9, 9:4, 4:'table', 12:2, 2:3, 3:16, 16:11, 11:7, 7:6, 6:'table'}
    goal3.clear = {17:True, 15:True, 12:True}

    goal3.display()
    
    expected = [('unstack', 1, 12), ('putdown', 1), ('unstack', 19, 18), ('putdown', 19), ('unstack', 18, 17), ('putdown', 18), ('unstack', 17, 16), ('putdown', 17), ('unstack', 9, 8), ('putdown', 9), ('unstack', 8, 7), ('putdown', 8), ('unstack', 11, 10), ('stack', 11, 7), ('unstack', 10, 5), ('putdown', 10), ('unstack', 5, 4), ('putdown', 5), ('unstack', 4, 14), ('putdown', 4), ('pickup', 9), ('stack', 9, 4), ('pickup', 8), ('stack', 8, 9), ('unstack', 14, 15), ('putdown', 14), ('unstack', 16, 3), ('stack', 16, 11), ('unstack', 3, 2), ('stack', 3, 16), ('pickup', 2), ('stack', 2, 3), ('unstack', 12, 13), ('stack', 12, 2), ('pickup', 13), ('stack', 13, 8), ('pickup', 15), ('stack', 15, 13)] 

    plan = gtpyhop.find_plan(state3,[goal3])
    th.check_result(plan,expected)
    th.pause(do_pauses)

    print("\nRun find_plan on problem BW-rand-50 from the IPC-2011 distribution.\n")

     
    print("- Define initial state for problem IPC2011BWrand50:")
     
    IPC2011BWrand50 = gtpyhop.State('problem BW-rand-50')
    IPC2011BWrand50.pos = {    
            1:48, 2:33, 3:41, 4:37, 5:45, 6:16, 7:31, 8:28, 9:49,
            10:34, 11:15, 12:17, 13:20, 14:2, 15:44, 16:5, 17:32, 18:50, 19:30,
            20:22, 21:27, 22:38, 23:11, 24:'table', 25:46, 26:'table', 27:40, 28:43, 29:19,
            30:39, 31:29, 32:'table', 33:'table', 34:14, 35:36, 36:'table', 37:8, 38:9, 39:18,
            40:3, 41:35, 42:4, 43:24, 44:26, 45:47, 46:42, 47:1, 48:21, 49:25,
            50:6
            }
    IPC2011BWrand50.clear = {x:False for x in range(1,50)}
    IPC2011BWrand50.clear.update({7:True, 10:True, 12:True, 13:True, 23:True})
    IPC2011BWrand50.holding = {'hand':False}
     
    IPC2011BWrand50.display('Initial state is')

     
    IPC2011BWrand50Goal = gtpyhop.Multigoal('problem BW-rand-50')
    IPC2011BWrand50Goal.pos = {1:33, 3:40, 4:46, 5:21, 6:17, 7:37, 8:15, 9:41,  
            10:26, 11:23, 12:25, 13:47, 14:20, 15:19, 16:31, 17:39, 18:50, 19:1,
            20:45, 21:11, 23:43, 25:42, 26:36, 27:35, 28:29, 29:44,
            30:8, 31:9, 32:6, 33:10, 34:14, 35:2, 36:7, 37:32, 38:28,
            40:24, 41:38, 42:34, 43:12, 44:49, 45:4, 46:18, 47:30, 48:22,
            50:13}
    IPC2011BWrand50Goal.clear = {}
     
    IPC2011BWrand50Goal.display()

    expected = [('unstack', 7, 31), ('putdown', 7), ('unstack', 10, 34), ('putdown', 10), ('unstack', 12, 17), ('putdown', 12), ('unstack', 13, 20), ('putdown', 13), ('unstack', 17, 32), ('putdown', 17), ('unstack', 20, 22), ('putdown', 20), ('unstack', 22, 38), ('putdown', 22), ('unstack', 23, 11), ('putdown', 23), ('unstack', 11, 15), ('putdown', 11), ('unstack', 15, 44), ('putdown', 15), ('unstack', 31, 29), ('putdown', 31), ('unstack', 29, 19), ('putdown', 29), ('unstack', 19, 30), ('putdown', 19), ('unstack', 30, 39), ('putdown', 30), ('unstack', 39, 18), ('putdown', 39), ('pickup', 17), ('stack', 17, 39), ('unstack', 18, 50), ('putdown', 18), ('unstack', 34, 14), ('putdown', 34), ('unstack', 14, 2), ('putdown', 14), ('unstack', 2, 33), ('putdown', 2), ('unstack', 38, 9), ('putdown', 38), ('unstack', 9, 49), ('putdown', 9), ('unstack', 49, 25), ('putdown', 49), ('unstack', 44, 26), ('stack', 44, 49), ('pickup', 29), ('stack', 29, 44), ('unstack', 25, 46), ('putdown', 25), ('unstack', 46, 42), ('putdown', 46), ('unstack', 42, 4), ('putdown', 42), ('unstack', 4, 37), ('putdown', 4), ('unstack', 37, 8), ('putdown', 37), ('unstack', 8, 28), ('putdown', 8), ('unstack', 28, 43), ('stack', 28, 29), ('pickup', 38), ('stack', 38, 28), ('unstack', 43, 24), ('putdown', 43), ('unstack', 50, 6), ('putdown', 50), ('unstack', 6, 16), ('stack', 6, 17), ('pickup', 32), ('stack', 32, 6), ('pickup', 37), ('stack', 37, 32), ('pickup', 7), ('stack', 7, 37), ('unstack', 16, 5), ('putdown', 16), ('unstack', 5, 45), ('putdown', 5), ('unstack', 45, 47), ('putdown', 45), ('unstack', 47, 1), ('putdown', 47), ('unstack', 1, 48), ('putdown', 1), ('unstack', 48, 21), ('stack', 48, 22), ('unstack', 21, 27), ('putdown', 21), ('unstack', 27, 40), ('putdown', 27), ('unstack', 40, 3), ('stack', 40, 24), ('unstack', 3, 41), ('stack', 3, 40), ('unstack', 41, 35), ('stack', 41, 38), ('pickup', 9), ('stack', 9, 41), ('pickup', 31), ('stack', 31, 9), ('pickup', 16), ('stack', 16, 31), ('unstack', 35, 36), ('stack', 35, 2), ('pickup', 27), ('stack', 27, 35), ('pickup', 36), ('stack', 36, 7), ('pickup', 26), ('stack', 26, 36), ('pickup', 10), ('stack', 10, 26), ('pickup', 33), ('stack', 33, 10), ('pickup', 1), ('stack', 1, 33), ('pickup', 19), ('stack', 19, 1), ('pickup', 15), ('stack', 15, 19), ('pickup', 8), ('stack', 8, 15), ('pickup', 30), ('stack', 30, 8), ('pickup', 47), ('stack', 47, 30), ('pickup', 13), ('stack', 13, 47), ('pickup', 50), ('stack', 50, 13), ('pickup', 18), ('stack', 18, 50), ('pickup', 46), ('stack', 46, 18), ('pickup', 4), ('stack', 4, 46), ('pickup', 45), ('stack', 45, 4), ('pickup', 20), ('stack', 20, 45), ('pickup', 14), ('stack', 14, 20), ('pickup', 34), ('stack', 34, 14), ('pickup', 42), ('stack', 42, 34), ('pickup', 25), ('stack', 25, 42), ('pickup', 12), ('stack', 12, 25), ('pickup', 43), ('stack', 43, 12), ('pickup', 23), ('stack', 23, 43), ('pickup', 11), ('stack', 11, 23), ('pickup', 21), ('stack', 21, 11), ('pickup', 5), ('stack', 5, 21)] 
    
    plan = gtpyhop.find_plan(IPC2011BWrand50, [IPC2011BWrand50Goal])
    th.check_result(plan,expected)
    th.pause(do_pauses)

    print("""
Call run_lazy_lookahead on the following problem, with verbose=1:
""")

    state2.display(heading='Initial state is')
    goal2b.display(heading='Goal is')

    new_state = gtpyhop.run_lazy_lookahead(state2, [goal2b])

    th.pause(do_pauses)

    print("The goal should now be satisfied, so the planner should return an empty plan:\n")

    plan = gtpyhop.find_plan(new_state, [goal2b])
    th.check_result(plan,[])

    print("No more examples")


# It's tempting to make the following call to main() unconditional, to run the
# examples without making the user type an extra command. But if we do this
# and an error occurs while main() is executing, we get a situation in which
# the actions, methods, and examples files have been imported but the module
# hasn't been - which causes problems if we try to import the module again.

if __name__=="__main__":
    main()



"""
Examples file for blocks_goal_splitting.
-- Dana Nau <nau@umd.edu>, July 20, 2021
"""

# Uncomment this to use it in debugging:
# from IPython import embed
# from IPython.terminal.debugger import set_trace

import gtpyhop
import gtpyhop.test_harness as th   # code for use in paging and debugging


# We must declare the current domain before importing methods and actions.
# To make the code more portable, we don't hard-code the domain name, but
# instead use the name of the package.
the_domain = gtpyhop.Domain(__package__)

from .methods import *
from .actions import *

print('-----------------------------------------------------------------------')
print(f"Created '{gtpyhop.current_domain}'. To run the examples, type this:")
print(f'{the_domain.__name__}.main()')


#############     beginning of tests     ################

def main(do_pauses=True):
    """
    Run various examples.
    main() will pause occasionally to let you examine the output.
    main(False) will run straight through to the end, without stopping.
    """

    # If we've changed to some other domain, this will change us back.
    print(f"Changing current domain to {the_domain}, if it isn't that already.")
    gtpyhop.set_current_domain(gtpyhop.find_domain_by_name(gtpyhop.Domain(__package__).__name__))
 
    gtpyhop.print_domain()

    state1 = gtpyhop.State('state1')
    state1.pos={'a':'b', 'b':'table', 'c':'table'}
    state1.clear={'c':True, 'b':False,'a':True}
    state1.holding={'hand':False}

    state1.display('\nInitial state is')

    print("""
Below, both goal1a and goal1b specify that we want c on b, and b on a. 
However, goal1a also specifies that we want a on the table.
""")

    goal1a = gtpyhop.Multigoal('goal1a')
    goal1a.pos={'c':'b', 'b':'a', 'a':'table'}

    goal1a.display()

    goal1b = gtpyhop.Multigoal('goal1b')
    goal1b.pos={'c':'b', 'b':'a'}

    goal1b.display()

    ### goal1b omits some of the conditions of goal1a,
    ### but those conditions will need to be achieved anyway

    th.pause(do_pauses)

    print("""
Run GTPyhop with goal1a and goal1b, starting in state1. Both should produce the
same plan, but it won't be a very good plan, because m_split_multigoal doesn't know
how to choose the best order for achieving the goals.
""")

    state1.display("Initial state is")

    # Tell the test harness what answer to expect, so it can signal an error
    # if gtpyhop returns an incorrect answer. Checks like this have been very
    # very helpful for debugging both gtpyhop and the various example domains.

    expected = [('unstack', 'a', 'b'), ('putdown', 'a'), ('pickup', 'c'), ('stack', 'c', 'b'), ('unstack', 'c', 'b'), ('putdown', 'c'), ('pickup', 'b'), ('stack', 'b', 'a'), ('pickup', 'c'), ('stack', 'c', 'b')]

    plan1 = gtpyhop.find_plan(state1,[goal1a])
    th.check_result(plan1,expected)

    plan2 = gtpyhop.find_plan(state1,[goal1b])
    th.check_result(plan2,expected)
    th.pause(do_pauses)


    print("""
Run GTPyhop on two more planning problems. Like before, goal2a omits some
of the conditions in goal2a, but both goals should produce the same plan.
""")

    state2 = gtpyhop.State('state2')
    state2.pos={'a':'c', 'b':'d', 'c':'table', 'd':'table'}
    state2.clear={'a':True, 'c':False,'b':True, 'd':False}
    state2.holding={'hand':False}

    state2.display('The initial state is')
    
    goal2a = gtpyhop.Multigoal('goal2a')
    goal2a.pos={'b':'c', 'a':'d', 'c':'table', 'd':'table'}
    goal2a.clear={'a':True, 'c':False,'b':True, 'd':False}
    goal2a.holding={'hand':False}

    goal2a.display()
    
    goal2b = gtpyhop.Multigoal('goal2b')
    goal2b.pos={'b':'c', 'a':'d'}

    goal2b.display()
    
    ### goal2b omits some of the conditions of goal2a,
    ### but those conditions will need to be achieved anyway.

    expected = [('unstack', 'a', 'c'), ('putdown', 'a'), ('unstack', 'b', 'd'), ('stack', 'b', 'c'), ('pickup', 'a'), ('stack', 'a', 'd')] 

    plan1 = gtpyhop.find_plan(state2,[goal2a])
    th.check_result(plan1,expected)

    plan2 = gtpyhop.find_plan(state2,[goal2b])
    th.check_result(plan2,expected)
    th.pause(do_pauses)


    # In some Python installations, the default recursion limit is too small to
    # run gtpyhop on the next planning problem. If so, we increase it.

    import sys
    reclimit = sys.getrecursionlimit()
    if reclimit < 1200:
        print(f"Increase Python's recursion limit from {reclimit} to 1200""", end='')
        sys.setrecursionlimit(1200)
        print(" ... done.")

    print("""
Define problem bw_large_d from the SHOP distribution. 
""")

    state3 = gtpyhop.State('state3')
    state3.pos = {1:12, 12:13, 13:'table', 11:10, 10:5, 5:4, 4:14, 14:15, 15:'table', 9:8, 8:7, 7:6, 6:'table', 19:18, 18:17, 17:16, 16:3, 3:2, 2:'table'}
    state3.clear = {x:False for x in range(1,20)}
    state3.clear.update({1:True, 11:True, 9:True, 19:True})
    state3.holding={'hand':False}

    state3.display('The initial state is')
    
    goal3 = gtpyhop.Multigoal('goal3')
    goal3.pos = {15:13, 13:8, 8:9, 9:4, 4:'table', 12:2, 2:3, 3:16, 16:11, 11:7, 7:6, 6:'table'}
    goal3.clear = {17:True, 15:True, 12:True}

    goal3.display()
    
    expected = [('unstack', 1, 12), ('putdown', 1), ('unstack', 12, 13), ('putdown', 12), ('unstack', 11, 10), ('putdown', 11), ('unstack', 10, 5), ('putdown', 10), ('unstack', 5, 4), ('putdown', 5), ('unstack', 4, 14), ('putdown', 4), ('unstack', 14, 15), ('putdown', 14), ('pickup', 15), ('stack', 15, 13), ('unstack', 9, 8), ('putdown', 9), ('unstack', 15, 13), ('putdown', 15), ('pickup', 13), ('stack', 13, 8), ('unstack', 13, 8), ('putdown', 13), ('unstack', 8, 7), ('stack', 8, 9), ('unstack', 8, 9), ('putdown', 8), ('pickup', 9), ('stack', 9, 4), ('unstack', 19, 18), ('putdown', 19), ('unstack', 18, 17), ('putdown', 18), ('unstack', 17, 16), ('putdown', 17), ('unstack', 16, 3), ('putdown', 16), ('unstack', 3, 2), ('putdown', 3), ('pickup', 12), ('stack', 12, 2), ('unstack', 12, 2), ('putdown', 12), ('pickup', 2), ('stack', 2, 3), ('unstack', 2, 3), ('putdown', 2), ('pickup', 3), ('stack', 3, 16), ('unstack', 3, 16), ('putdown', 3), ('pickup', 16), ('stack', 16, 11), ('unstack', 16, 11), ('putdown', 16), ('pickup', 11), ('stack', 11, 7), ('pickup', 15), ('stack', 15, 13), ('unstack', 15, 13), ('putdown', 15), ('pickup', 13), ('stack', 13, 8), ('unstack', 13, 8), ('putdown', 13), ('pickup', 8), ('stack', 8, 9), ('pickup', 12), ('stack', 12, 2), ('unstack', 12, 2), ('putdown', 12), ('pickup', 2), ('stack', 2, 3), ('unstack', 2, 3), ('putdown', 2), ('pickup', 3), ('stack', 3, 16), ('unstack', 3, 16), ('putdown', 3), ('pickup', 16), ('stack', 16, 11), ('pickup', 15), ('stack', 15, 13), ('unstack', 15, 13), ('putdown', 15), ('pickup', 13), ('stack', 13, 8), ('pickup', 12), ('stack', 12, 2), ('unstack', 12, 2), ('putdown', 12), ('pickup', 2), ('stack', 2, 3), ('unstack', 2, 3), ('putdown', 2), ('pickup', 3), ('stack', 3, 16), ('pickup', 15), ('stack', 15, 13), ('pickup', 12), ('stack', 12, 2), ('unstack', 12, 2), ('putdown', 12), ('pickup', 2), ('stack', 2, 3), ('pickup', 12), ('stack', 12, 2)]
    th.pause(do_pauses)


    print("""
Run gtpyhop on bw_large_d. It will find a long plan, because m_split_multigoal
doesn't know how to choose the best order for achieving the goals.
""")

    plan = gtpyhop.find_plan(state3,[goal3])
    th.check_result(plan,expected)
    th.pause(do_pauses)


    print("""
Call run_lazy_lookahead on the same problem:
""")

    new_state = gtpyhop.run_lazy_lookahead(state3, [goal3])

    print("The goal should now be satisfied, so the planner should return an empty plan:\n")

    plan = gtpyhop.find_plan(new_state, [goal3])
    th.check_result(plan,[])

    print("No more examples")


# It's tempting to make the following call to main() unconditional, to run the
# examples without making the user type an extra command. But if we do this
# and an error occurs while main() is executing, we get a situation in which
# the actions, methods, and examples files have been imported but the module
# hasn't been - which causes problems if we try to import the module again.

if __name__=="__main__":
    main()



"""
The following code runs GTPyhop on all but one of the example domains, to
see whether they run without error and return correct answers. The
2nd-to-last line imports simple_htn_acting_error but doesn't run it,
because running it is *supposed* to cause an error.

-- Dana Nau <nau@umd.edu>, July 20, 2021
"""

def main():
    """
    Run all the regression tests.
    """
    # The argument False tells the test harness not to stop for user input
    from gtpyhop.examples import simple_htn; simple_htn.main(False)
    from gtpyhop.examples import simple_hgn; simple_hgn.main(False)
    
    # skip testing the recursive backtracking when planning is iterative
    from src.gtpyhop.main import get_recursive_planning
    if get_recursive_planning():
         from gtpyhop.examples import backtracking_htn; backtracking_htn.main(False)
    
    from gtpyhop.examples import logistics_hgn; logistics_hgn.main(False)
   
    # Python recursion limit is set to 1000 by default which is:
    #  - not enough for both blocks_gtn and blocks_hgn examples,
    #  - enough for blocks_goal_splitting, and blocks_htn examples.
    if get_recursive_planning():
        import sys
        current_recursion_limit = sys.getrecursionlimit()
        blocks_gtn_recursion_limit = 2000
        # Increase recursion limit for blocks_gtn test if needed
        if current_recursion_limit < blocks_gtn_recursion_limit:
            sys.setrecursionlimit(blocks_gtn_recursion_limit)  # Increase recursion limit for the next test
            print(f"Recursion limit set to {blocks_gtn_recursion_limit} for blocks_gtn test.")
       
    # With an appropriate recursion limit, next 2 tests should run without error
    # Of course, if we're running with iterative planning, we don't need to increase the recursion limit.
    from gtpyhop.examples import blocks_gtn; blocks_gtn.main(False)
    from gtpyhop.examples import blocks_hgn; blocks_hgn.main(False)

    # Restore the original recursion limit after the test if needed
    if get_recursive_planning():
        if current_recursion_limit < blocks_gtn_recursion_limit:
            sys.setrecursionlimit(current_recursion_limit)
            print(f"Recursion limit restored to {current_recursion_limit} after blocks_gtn test.")

    from gtpyhop.examples import blocks_goal_splitting; blocks_goal_splitting.main(False)  
    from gtpyhop.examples import blocks_htn; blocks_htn.main(False)
    from gtpyhop.examples import pyhop_simple_travel_example
    from gtpyhop.examples import simple_htn_acting_error
    print('\nFinished without error.')



"""
Version optimisée du simple_htn.py avec gestion mémoire contrôlée
Utilise unittest pour gérer l'allocation et désallocation des instances State
-- Optimisation par Claude, basée sur le code de Dana Nau
"""

import unittest
import weakref
import gc
import sys
from typing import List, Dict, Any, Optional, Set
from contextlib import contextmanager
import gtpyhop
import random
import gtpyhop.test_harness as th


# Configuration du domaine
domain_name = __name__
the_domain = gtpyhop.Domain(domain_name)


class StateManager:
    """Gestionnaire centralisé pour les instances State avec tracking mémoire"""
    
    def __init__(self):
        self._state_registry: Set[weakref.ref] = set()
        self._creation_count = 0
        self._cleanup_count = 0
    
    def register_state(self, state: gtpyhop.State) -> gtpyhop.State:
        """Enregistre une nouvelle instance State"""
        def cleanup_callback(ref):
            self._state_registry.discard(ref)
            self._cleanup_count += 1
        
        ref = weakref.ref(state, cleanup_callback)
        self._state_registry.add(ref)
        self._creation_count += 1
        return state
    
    def get_active_states_count(self) -> int:
        """Retourne le nombre d'instances State actives"""
        # Nettoie les références mortes
        self._state_registry = {ref for ref in self._state_registry if ref() is not None}
        return len(self._state_registry)
    
    def force_cleanup(self):
        """Force le nettoyage des instances State"""
        gc.collect()
        self._state_registry = {ref for ref in self._state_registry if ref() is not None}
    
    def get_stats(self) -> Dict[str, int]:
        """Retourne les statistiques d'allocation"""
        return {
            'created': self._creation_count,
            'cleaned': self._cleanup_count,
            'active': self.get_active_states_count()
        }
    
    def reset_stats(self):
        """Remet à zéro les statistiques"""
        self._creation_count = 0
        self._cleanup_count = 0
        self.force_cleanup()


# Instance globale du gestionnaire d'états
state_manager = StateManager()


class ManagedState(gtpyhop.State):
    """Version de State avec gestion mémoire automatique"""
    
    def __init__(self, name: str = 'managed_state'):
        super().__init__(name)
        self.name = name  # Initialize the name attribute
        state_manager.register_state(self)
    
    def copy(self) -> 'ManagedState':
        """Copie gérée avec tracking automatique"""
        new_state = ManagedState(f"{self.name}_copy")
        
        # Copie des attributs de base
        for attr_name in dir(self):
            if not attr_name.startswith('_') and attr_name != 'name':
                attr_value = getattr(self, attr_name)
                if hasattr(attr_value, 'copy'):
                    setattr(new_state, attr_name, attr_value.copy())
                elif isinstance(attr_value, dict):
                    setattr(new_state, attr_name, attr_value.copy())
                elif isinstance(attr_value, list):
                    setattr(new_state, attr_name, attr_value.copy())
                else:
                    setattr(new_state, attr_name, attr_value)
        
        return new_state


################################################################################
# États et relations rigides

rigid = ManagedState('rigid_relations')
rigid.types = {
    'person':   ['alice', 'bob'],
    'location': ['home_a', 'home_b', 'park', 'station'],
    'taxi':     ['taxi1', 'taxi2']
}
rigid.dist = {
    ('home_a', 'park'): 8,    ('home_b', 'park'): 2,
    ('station', 'home_a'): 1, ('station', 'home_b'): 7,
    ('home_a', 'home_b'): 7,  ('station', 'park'): 9
}


def create_initial_state() -> ManagedState:
    """Factory pour créer l'état initial"""
    state = ManagedState('initial_state')
    state.loc = {'alice': 'home_a', 'bob': 'home_b', 'taxi1': 'park', 'taxi2': 'station'}
    state.cash = {'alice': 20, 'bob': 15}
    state.owe = {'alice': 0, 'bob': 0}
    return state


###############################################################################
# Fonctions utilitaires

def taxi_rate(dist: float) -> float:
    """Calcule le tarif du taxi"""
    return 1.5 + 0.5 * dist


def distance(x: str, y: str) -> Optional[float]:
    """Calcule la distance entre deux locations"""
    return rigid.dist.get((x, y)) or rigid.dist.get((y, x))


def is_a(variable: str, type_name: str) -> bool:
    """Vérifie le type d'une variable"""
    return variable in rigid.types[type_name]


###############################################################################
# Actions

def walk(state: ManagedState, p: str, x: str, y: str) -> Optional[ManagedState]:
    """Action de marche"""
    if is_a(p, 'person') and is_a(x, 'location') and is_a(y, 'location') and x != y:
        if state.loc[p] == x:
            state.loc[p] = y
            return state
    return None


def call_taxi(state: ManagedState, p: str, x: str) -> Optional[ManagedState]:
    """Action d'appel de taxi"""
    if is_a(p, 'person') and is_a(x, 'location'):
        state.loc['taxi1'] = x
        state.loc[p] = 'taxi1'
        return state
    return None


def ride_taxi(state: ManagedState, p: str, y: str) -> Optional[ManagedState]:
    """Action de voyage en taxi"""
    if is_a(p, 'person') and is_a(state.loc[p], 'taxi') and is_a(y, 'location'):
        taxi = state.loc[p]
        x = state.loc[taxi]
        if is_a(x, 'location') and x != y:
            state.loc[taxi] = y
            state.owe[p] = taxi_rate(distance(x, y))
            return state
    return None


def pay_driver(state: ManagedState, p: str, y: str) -> Optional[ManagedState]:
    """Action de paiement du chauffeur"""
    if is_a(p, 'person'):
        if state.cash[p] >= state.owe[p]:
            state.cash[p] = state.cash[p] - state.owe[p]
            state.owe[p] = 0
            state.loc[p] = y
            return state
    return None


# Déclaration des actions
gtpyhop.declare_actions(walk, call_taxi, ride_taxi, pay_driver)


###############################################################################
# Commandes avec gestion d'erreur

def c_walk(state: ManagedState, p: str, x: str, y: str) -> Optional[ManagedState]:
    """Commande de marche"""
    return walk(state, p, x, y)


def c_call_taxi(state: ManagedState, p: str, x: str) -> Optional[ManagedState]:
    """Commande d'appel de taxi (avec probabilité d'échec)"""
    if is_a(p, 'person') and is_a(x, 'location'):
        if random.randrange(2) > 0:
            state.loc['taxi1'] = x
            state.loc[p] = 'taxi1'
            print('Action> c_call_taxi succeeded. Pr = 1/2')
            return state
        else:
            print('Action> c_call_taxi failed. Pr = 1/2')
            return None
    return None


def c_ride_taxi(state: ManagedState, p: str, y: str) -> Optional[ManagedState]:
    """Commande de voyage en taxi"""
    return ride_taxi(state, p, y)


def c_pay_driver(state: ManagedState, p: str, y: str) -> Optional[ManagedState]:
    """Commande de paiement du chauffeur"""
    return pay_driver(state, p, y)


# Déclaration des commandes
gtpyhop.declare_commands(c_walk, c_call_taxi, c_ride_taxi, c_pay_driver)


###############################################################################
# Méthodes

def do_nothing(state: ManagedState, p: str, y: str) -> Optional[List]:
    """Méthode pour ne rien faire si déjà à destination"""
    if is_a(p, 'person') and is_a(y, 'location'):
        if state.loc[p] == y:
            return []
    return None


def travel_by_foot(state: ManagedState, p: str, y: str) -> Optional[List]:
    """Méthode de voyage à pied"""
    if is_a(p, 'person') and is_a(y, 'location'):
        x = state.loc[p]
        if x != y and distance(x, y) <= 2:
            return [('walk', p, x, y)]
    return None


def travel_by_taxi(state: ManagedState, p: str, y: str) -> Optional[List]:
    """Méthode de voyage en taxi"""
    if is_a(p, 'person') and is_a(y, 'location'):
        x = state.loc[p]
        if x != y and state.cash[p] >= taxi_rate(distance(x, y)):
            return [('call_taxi', p, x), ('ride_taxi', p, y), ('pay_driver', p, y)]
    return None


# Déclaration des méthodes
gtpyhop.declare_task_methods('travel', do_nothing, travel_by_foot, travel_by_taxi)


###############################################################################
# Tests unitaires avec gestion mémoire

class HTNPlanningTestCase(unittest.TestCase):
    """Classe de test avec gestion mémoire optimisée"""
    
    @classmethod
    def setUpClass(cls):
        """Configuration une fois pour toute la classe"""
        gtpyhop.current_domain = the_domain
        print(f"\n=== Début des tests pour le domaine '{domain_name}' ===")
        
    @classmethod
    def tearDownClass(cls):
        """Nettoyage final après tous les tests"""
        state_manager.force_cleanup()
        final_stats = state_manager.get_stats()
        print(f"\n=== Statistiques finales ===")
        print(f"États créés: {final_stats['created']}")
        print(f"États nettoyés: {final_stats['cleaned']}")
        print(f"États actifs: {final_stats['active']}")
        
        if final_stats['active'] > 0:
            print(f"⚠️  Attention: {final_stats['active']} états non nettoyés")
        else:
            print("✅ Tous les états ont été correctement nettoyés")
    
    def setUp(self):
        """Configuration avant chaque test"""
        self.initial_stats = state_manager.get_stats()
        self.test_states = []
        
    def tearDown(self):
        """Nettoyage après chaque test"""
        # Libère explicitement les références locales
        for state in self.test_states:
            del state
        self.test_states.clear()
        
        # Force le garbage collection
        state_manager.force_cleanup()
        
        # Vérifie les fuites mémoire
        current_stats = state_manager.get_stats()
        created_in_test = current_stats['created'] - self.initial_stats['created']
        cleaned_in_test = current_stats['cleaned'] - self.initial_stats['cleaned']
        
        if created_in_test > 0:
            print(f"Test {self._testMethodName}: {created_in_test} états créés, "
                  f"{cleaned_in_test} nettoyés")
    
    @contextmanager
    def memory_tracking(self, test_name: str):
        """Context manager pour tracker l'utilisation mémoire"""
        before = state_manager.get_stats()
        try:
            yield
        finally:
            state_manager.force_cleanup()
            after = state_manager.get_stats()
            created = after['created'] - before['created']
            cleaned = after['cleaned'] - before['cleaned']
            active = after['active'] - before['active']
            
            if created > 0:
                print(f"  {test_name}: +{created} créés, +{cleaned} nettoyés, "
                      f"{active:+d} actifs")
    
    def create_test_state(self, name: str = None) -> ManagedState:
        """Factory pour créer un état de test géré"""
        state = create_initial_state()
        if name:
            state.name = name
        self.test_states.append(state)
        return state
    
    def test_basic_planning(self):
        """Test de planification de base"""
        with self.memory_tracking("test_basic_planning"):
            state = self.create_test_state("basic_test")
            
            gtpyhop.verbose = 0
            result = gtpyhop.find_plan(state, [('travel', 'alice', 'park')])
            
            expected = [('call_taxi', 'alice', 'home_a'), 
                       ('ride_taxi', 'alice', 'park'), 
                       ('pay_driver', 'alice', 'park')]
            
            self.assertEqual(result, expected)
    
    def test_multiple_travelers(self):
        """Test avec plusieurs voyageurs"""
        with self.memory_tracking("test_multiple_travelers"):
            state = self.create_test_state("multiple_test")
            
            gtpyhop.verbose = 0
            result = gtpyhop.find_plan(state, [('travel', 'alice', 'park'), 
                                               ('travel', 'bob', 'park')])
            
            expected = [('call_taxi', 'alice', 'home_a'),
                       ('ride_taxi', 'alice', 'park'),
                       ('pay_driver', 'alice', 'park'),
                       ('walk', 'bob', 'home_b', 'park')]
            
            self.assertEqual(result, expected)
    
    def test_state_copying(self):
        """Test de copie d'états"""
        with self.memory_tracking("test_state_copying"):
            original = self.create_test_state("original")
            original.cash['alice'] = 100
            
            # Test de copie
            copy1 = original.copy()
            copy2 = original.copy()
            
            # Ajoute les copies à la liste de gestion
            self.test_states.extend([copy1, copy2])
            
            # Vérifie que les copies sont indépendantes
            copy1.cash['alice'] = 50
            copy2.cash['alice'] = 25
            
            self.assertEqual(original.cash['alice'], 100)
            self.assertEqual(copy1.cash['alice'], 50)
            self.assertEqual(copy2.cash['alice'], 25)
    
    def test_no_solution(self):
        """Test avec problème sans solution"""
        with self.memory_tracking("test_no_solution"):
            state = self.create_test_state("no_solution")
            # Enlève tout l'argent
            state.cash = {'alice': 0, 'bob': 0}
            
            gtpyhop.verbose = 0
            result = gtpyhop.find_plan(state, [('travel', 'alice', 'park')])
            
            # Aucune solution possible
            self.assertIsNone(result)
    
    def test_already_at_destination(self):
        """Test quand déjà à destination"""
        with self.memory_tracking("test_already_at_destination"):
            state = self.create_test_state("already_there")
            state.loc['alice'] = 'park'
            
            gtpyhop.verbose = 0
            result = gtpyhop.find_plan(state, [('travel', 'alice', 'park')])
            
            # Plan vide car déjà à destination
            self.assertEqual(result, [])
    
    def test_memory_pressure(self):
        """Test de pression mémoire avec nombreuses copies"""
        with self.memory_tracking("test_memory_pressure"):
            base_state = self.create_test_state("base")
            
            # Crée de nombreuses copies
            copies = []
            for i in range(100):
                copy = base_state.copy()
                copies.append(copy)
                
            # Ajoute à la gestion
            self.test_states.extend(copies)
            
            # Vérifie que toutes les copies sont distinctes
            for i, copy in enumerate(copies):
                copy.cash['alice'] = i
                
            # Vérifie l'indépendance
            values = [copy.cash['alice'] for copy in copies]
            self.assertEqual(values, list(range(100)))
    
    def test_deep_copy_integrity(self):
        """Test de l'intégrité des copies profondes"""
        with self.memory_tracking("test_deep_copy_integrity"):
            original = self.create_test_state("deep_copy_test")
            
            # Modifie les structures imbriquées
            original.complex_data = {
                'nested': {'level1': {'level2': 'value'}},
                'list_data': [1, 2, [3, 4]]
            }
            
            # Crée une copie
            copy = original.copy()
            self.test_states.append(copy)
            
            # Modifie l'original
            original.complex_data['nested']['level1']['level2'] = 'modified'
            original.complex_data['list_data'][2][0] = 999
            
            # Vérifie que la copie n'est pas affectée
            self.assertEqual(copy.complex_data['nested']['level1']['level2'], 'value')
            self.assertEqual(copy.complex_data['list_data'][2][0], 3)
    
    def test_concurrent_planning(self):
        """Test de planification concurrente"""
        with self.memory_tracking("test_concurrent_planning"):
            states = []
            plans = []
            
            # Crée plusieurs états avec des conditions différentes
            for i in range(10):
                state = self.create_test_state(f"concurrent_{i}")
                state.cash['alice'] = 10 + i * 5  # Cash variable
                states.append(state)
            
            # Planifie pour chaque état
            gtpyhop.verbose = 0
            for state in states:
                plan = gtpyhop.find_plan(state, [('travel', 'alice', 'park')])
                plans.append(plan)
            
            # Vérifie que toutes les planifications ont réussi
            for plan in plans:
                self.assertIsNotNone(plan)
                self.assertGreater(len(plan), 0)


class HTNMemoryLeakTestCase(unittest.TestCase):
    """Tests spécifiques pour détecter les fuites mémoire"""
    
    def setUp(self):
        """Configuration avant chaque test"""
        state_manager.reset_stats()
        self.initial_memory = state_manager.get_stats()
    
    def tearDown(self):
        """Vérification après chaque test"""
        state_manager.force_cleanup()
        final_memory = state_manager.get_stats()
        
        # Vérifie qu'il n'y a pas de fuites
        if final_memory['active'] > 0:
            print(f"⚠️  Fuite détectée: {final_memory['active']} états actifs")
    
    def test_massive_state_creation(self):
        """Test de création massive d'états"""
        states = []
        
        try:
            for i in range(1000):
                state = create_initial_state()
                state.test_id = i
                states.append(state)
            
            # Vérifie que tous les états sont créés
            self.assertEqual(len(states), 1000)
            
            # Vérifie l'unicité
            ids = [state.test_id for state in states]
            self.assertEqual(len(set(ids)), 1000)
            
        finally:
            # Nettoyage explicite
            for state in states:
                del state
            del states
            
            # Force le garbage collection
            state_manager.force_cleanup()
    
    def test_recursive_copying(self):
        """Test de copie récursive"""
        def create_copy_chain(state, depth):
            if depth <= 0:
                return []
            
            copy = state.copy()
            copy.depth = depth
            copies = [copy]
            copies.extend(create_copy_chain(copy, depth - 1))
            return copies
        
        original = create_initial_state()
        
        try:
            # Crée une chaîne de copies
            copies = create_copy_chain(original, 50)
            
            # Vérifie la chaîne
            self.assertEqual(len(copies), 50)
            
            # Vérifie que chaque copie a la bonne profondeur
            for i, copy in enumerate(copies):
                expected_depth = 50 - i
                self.assertEqual(copy.depth, expected_depth)
                
        finally:
            # Nettoyage
            for copy in copies:
                del copy
            del copies
            del original
            
            state_manager.force_cleanup()


def run_performance_benchmark():
    """Benchmark de performance avec gestion mémoire"""
    print("\n=== Benchmark de performance ===")
    
    import time
    
    # Réinitialise les statistiques
    state_manager.reset_stats()
    
    start_time = time.time()
    
    # Test de création/copie massive
    base_state = create_initial_state()
    states = []
    
    for i in range(1000):
        if i % 100 == 0:
            print(f"  Création: {i}/1000")
        
        copy = base_state.copy()
        states.append(copy)
        
        # Planification simple
        gtpyhop.verbose = 0
        gtpyhop.find_plan(copy, [('travel', 'alice', 'park')])
    
    end_time = time.time()
    
    # Statistiques finales
    stats = state_manager.get_stats()
    
    print(f"\nRésultats du benchmark:")
    print(f"  Temps total: {end_time - start_time:.2f}s")
    print(f"  États créés: {stats['created']}")
    print(f"  États/seconde: {stats['created'] / (end_time - start_time):.1f}")
    print(f"  États actifs: {stats['active']}")
    
    # Nettoyage final
    del states
    del base_state
    state_manager.force_cleanup()
    
    final_stats = state_manager.get_stats()
    print(f"  États après nettoyage: {final_stats['active']}")


def run_memory_stress_test():
    """Test de stress mémoire"""
    print("\n=== Test de stress mémoire ===")
    
    import psutil
    import os
    
    process = psutil.Process(os.getpid())
    
    # Mémoire initiale
    initial_memory = process.memory_info().rss / 1024 / 1024  # MB
    print(f"Mémoire initiale: {initial_memory:.1f} MB")
    
    # Réinitialise les statistiques
    state_manager.reset_stats()
    
    # Crée de nombreux états de façon cyclique
    for cycle in range(10):
        print(f"  Cycle {cycle + 1}/10")
        
        # Crée des états
        states = []
        for i in range(500):
            state = create_initial_state()
            state.cycle = cycle
            state.index = i
            states.append(state)
        
        # Effectue des planifications
        gtpyhop.verbose = 0
        for state in states[:100]:  # Limite pour éviter la surcharge
            gtpyhop.find_plan(state, [('travel', 'alice', 'park')])
        
        # Libère les états
        del states
        
        # Force le nettoyage
        state_manager.force_cleanup()
        
        # Vérifie la mémoire
        current_memory = process.memory_info().rss / 1024 / 1024  # MB
        print(f"    Mémoire après cycle {cycle + 1}: {current_memory:.1f} MB")
    
    # Statistiques finales
    final_stats = state_manager.get_stats()
    final_memory = process.memory_info().rss / 1024 / 1024  # MB
    
    print(f"\nRésultats du stress test:")
    print(f"  Mémoire finale: {final_memory:.1f} MB")
    print(f"  Augmentation: {final_memory - initial_memory:.1f} MB")
    print(f"  États créés: {final_stats['created']}")
    print(f"  États actifs: {final_stats['active']}")
    
    if final_stats['active'] == 0:
        print("✅ Aucune fuite mémoire détectée")
    else:
        print(f"⚠️  {final_stats['active']} états non nettoyés")


def main():
    """Fonction principale pour exécuter les tests"""
    print(f"=== Démarrage des tests HTN avec gestion mémoire ===")
    print(f"Domaine: {domain_name}")
    print(f"Version Python: {sys.version}")
    
    # Exécute les tests unitaires
    print("\n1. Tests unitaires de base:")
    unittest.main(verbosity=2, exit=False, argv=[''])
    
    # Exécute le benchmark
    print("\n2. Benchmark de performance:")
    run_performance_benchmark()
    
    # Exécute le test de stress mémoire
    print("\n3. Test de stress mémoire:")
    run_memory_stress_test()
    
    print("\n=== Fin des tests ===")


if __name__ == '__main__':
    main()
